@@ -1,5 +1,6 @@
 from __future__ import annotations
 
+import dataclasses
 import os
 from functools import cached_property
 from typing import TYPE_CHECKING, Callable
@@ -9,10 +10,10 @@ from packaging.version import InvalidVersion
 from resolvelib import AbstractProvider, RequirementsConflicted
 from resolvelib.resolvers import Criterion
 
-from pdm.exceptions import InvalidPyVersion, RequirementError
+from pdm.exceptions import CandidateNotFound, InvalidPyVersion, RequirementError
 from pdm.models.candidates import Candidate
 from pdm.models.repositories import LockedRepository
-from pdm.models.requirements import FileRequirement, Requirement, parse_requirement, strip_extras
+from pdm.models.requirements import FileRequirement, Requirement, VcsRequirement, parse_requirement, strip_extras
 from pdm.models.specifiers import PySpecSet
 from pdm.resolver.python import PythonCandidate, PythonRequirement, find_python_matches, is_python_satisfied_by
 from pdm.termui import logger
@@ -25,13 +26,14 @@ from pdm.utils import (
 )
 
 if TYPE_CHECKING:
-    from typing import Any, Iterable, Iterator, Mapping, Sequence, TypeVar
+    from typing import Iterable, Iterator, Mapping, Sequence, TypeVar
 
     from resolvelib.resolvers import RequirementInformation
 
     from pdm._types import Comparable
-    from pdm.models.repositories import BaseRepository
+    from pdm.models.repositories import BaseRepository, LockedRepository
     from pdm.models.requirements import Requirement
+    from pdm.models.working_set import WorkingSet
 
     ProviderT = TypeVar("ProviderT", bound="type[BaseProvider]")
 
@@ -60,7 +62,7 @@ class BaseProvider(AbstractProvider[Requirement, Candidate, str]):
         overrides: dict[str, str] | None = None,
         direct_minimal_versions: bool = False,
         *,
-        locked_candidates: dict[str, list[Candidate]],
+        locked_repository: LockedRepository | None = None,
     ) -> None:
         if overrides is not None:  # pragma: no cover
             deprecation_warning(
@@ -76,7 +78,7 @@ class BaseProvider(AbstractProvider[Requirement, Candidate, str]):
         self.fetched_dependencies: dict[tuple[str, str | None], list[Requirement]] = {}
         self.excludes = {normalize_name(k) for k in project.pyproject.resolution.get("excludes", [])}
         self.direct_minimal_versions = direct_minimal_versions
-        self.locked_candidates = locked_candidates
+        self.locked_repository = locked_repository
 
     def requirement_preference(self, requirement: Requirement) -> Comparable:
         """Return the preference of a requirement to find candidates.
@@ -131,6 +133,10 @@ class BaseProvider(AbstractProvider[Requirement, Candidate, str]):
             identifier,
         )
 
+    @cached_property
+    def locked_candidates(self) -> dict[str, list[Candidate]]:
+        return self.locked_repository.all_candidates if self.locked_repository else {}
+
     @cached_property
     def overrides(self) -> dict[str, Requirement]:
         """A mapping of package name to the requirement for overriding."""
@@ -231,8 +237,11 @@ class BaseProvider(AbstractProvider[Requirement, Candidate, str]):
                 return (c for c in candidates if c not in incompat)
             elif identifier in self.overrides:
                 return iter(self._find_candidates(self.overrides[identifier]))
-            elif (name := strip_extras(identifier)[0]) in self.overrides:
-                return iter(self._find_candidates(self.overrides[name]))
+            else:
+                name, extras = strip_extras(identifier)
+                if name in self.overrides:
+                    req = dataclasses.replace(self.overrides[name], extras=extras)
+                    return iter(self._find_candidates(req))
             reqs = list(requirements[identifier])
             if not reqs:
                 return iter(())
@@ -286,17 +295,31 @@ class BaseProvider(AbstractProvider[Requirement, Candidate, str]):
         )
         return requirement.specifier.contains(version, allow_prereleases)
 
+    def _get_dependencies_from_repository(self, candidate: Candidate) -> tuple[list[Requirement], PySpecSet, str]:
+        return self.repository.get_dependencies(candidate)
+
     def get_dependencies(self, candidate: Candidate) -> list[Requirement]:
         if isinstance(candidate, PythonCandidate):
             return []
         try:
-            deps, requires_python, _ = self.repository.get_dependencies(candidate)
+            deps, requires_python, _ = self._get_dependencies_from_repository(candidate)
         except (RequirementError, InvalidPyVersion, InvalidSpecifier) as e:
             # When the metadata is invalid, skip this candidate by marking it as conflicting.
             # Here we pass an empty criterion so it doesn't provide any info to the resolution.
             logger.error("Invalid metadata in %s: %s", candidate, e)
             raise RequirementsConflicted(Criterion([], [], [])) from None
 
+        if candidate.req.extras:
+            # XXX: If the requirement has extras, add the original candidate
+            # (without extras) as its dependency. This ensures the same package with
+            # different extras resolve to the same version.
+            self_req = dataclasses.replace(
+                candidate.req.as_pinned_version(candidate.version),
+                extras=None,
+                marker=None,
+            )
+            if self_req not in deps:
+                deps.insert(0, self_req)
         self.fetched_dependencies[candidate.dep_key] = deps[:]
         # Filter out incompatible dependencies(e.g. functools32) early so that
         # we don't get errors when building wheels.
@@ -339,8 +362,23 @@ class ReusePinProvider(BaseProvider):
     where already-pinned candidates in lockfile should be preferred.
     """
 
-    def __init__(self, *args: Any, tracked_names: Iterable[str], **kwargs: Any) -> None:
-        super().__init__(*args, **kwargs)
+    def __init__(
+        self,
+        repository: BaseRepository,
+        allow_prereleases: bool | None = None,
+        overrides: dict[str, str] | None = None,
+        direct_minimal_versions: bool = False,
+        *,
+        locked_repository: LockedRepository | None = None,
+        tracked_names: Iterable[str],
+    ) -> None:
+        super().__init__(
+            repository=repository,
+            allow_prereleases=allow_prereleases,
+            overrides=overrides,
+            direct_minimal_versions=direct_minimal_versions,
+            locked_repository=locked_repository,
+        )
         self.tracked_names = set(tracked_names)
 
     def iter_reuse_candidates(self, identifier: str, requirement: Requirement | None) -> Iterable[Candidate]:
@@ -366,7 +404,8 @@ class ReusePinProvider(BaseProvider):
         def matches_gen() -> Iterator[Candidate]:
             requested_req = next(filter(lambda r: r.is_named, requirements[identifier]), None)
             for pin in self.iter_reuse_candidates(identifier, requested_req):
-                pin = pin.copy_with(min(requirements[identifier], key=self.requirement_preference))
+                if identifier not in self.overrides and pin.req.is_named:
+                    pin = pin.copy_with(min(requirements[identifier], key=self.requirement_preference))
                 incompat = list(incompatibilities[identifier])
                 pin._preferred = True  # type: ignore[attr-defined]
                 if pin not in incompat and all(self.is_satisfied_by(r, pin) for r in requirements[identifier]):
@@ -375,6 +414,17 @@ class ReusePinProvider(BaseProvider):
 
         return matches_gen
 
+    def _get_dependencies_from_repository(self, candidate: Candidate) -> tuple[list[Requirement], PySpecSet, str]:
+        is_stable_metadata = candidate.req.is_named or (
+            isinstance(candidate.req, VcsRequirement) and candidate.req.revision
+        )
+        if self.locked_repository is not None and is_stable_metadata:
+            try:
+                return self.locked_repository.get_dependencies(candidate)
+            except CandidateNotFound:
+                pass
+        return super()._get_dependencies_from_repository(candidate)
+
 
 @register_provider("eager")
 class EagerUpdateProvider(ReusePinProvider):
@@ -422,9 +472,9 @@ class EagerUpdateProvider(ReusePinProvider):
 class ReuseInstalledProvider(ReusePinProvider):
     """A provider that reuses installed packages if possible."""
 
-    def __init__(self, *args: Any, **kwargs: Any) -> None:
-        super().__init__(*args, **kwargs)
-        self.installed = self.repository.environment.get_working_set()
+    @cached_property
+    def installed(self) -> WorkingSet:
+        return self.repository.environment.get_working_set()
 
     def iter_reuse_candidates(self, identifier: str, requirement: Requirement | None) -> Iterable[Candidate]:
         key = strip_extras(identifier)[0]
